/************************************************************************
*									*
*	Device Driver for INMOS-B004/B008 compatible link interfaces	*
*									*
*	Copyright (C) 1993,94,95 by Christoph Niemann			*
*				 niemann@swt.ruhr-uni-bochum.de		*
*	based on /linux/kernel/drivers/char/lp.c			*
*									*
*	Version 0.8a, for Linux 1.2.X, March 1996.			*
*									*
*									*
*	Module support added by Mark Bowyer <mark@lindhard.demon.co.uk> *
*									*
*	The driver may be copied under the same conditions as the	*
*	Linux operating system						*
*									*
*************************************************************************/

/* Copyright (C) 1992 by Jim Weigand, Linus Torvalds, and Michael K. Johnson */


#ifndef __i386__
#error This driver requires the Intel architecture
#endif

/*
 * Do not define this symbol if you want to use the C-version of
 * the read/write functions (much slower)
 */
#define LINK_USE_ASM


#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/timer.h>
#include <linux/stat.h>
#include <linux/ioport.h>
#include <linux/linkage.h>

#include <asm/io.h>
#include <asm/segment.h>

#ifdef MODULE
#include "link.h"
#include <linux/module.h>
/*
 * NB. we must include the kernel idenfication string in to install the module.
 */
#include <linux/version.h>
char kernel_version[] = UTS_RELEASE;
#else
#include <linux/link.h>
#endif

/*
 * Define these symbols if you want to debug the code.
 */
#undef LINK_DEBUG 
#undef LINK_DEBUG_MORE

/* If your link interface does not work with the normal i/o-commands use
   the slower ones. But I think this should be no problem for any board
   that uses INMOS's C012 chip.
   Defining the symbol LINK_REALLY_SLOW_C012 will decrease the data-
   transfer-rate from about 400 KBytes/sec to 200 KBytes/sec
   (on my noname 486/33/16, assembler version).
*/
#undef LINK_REALLY_SLOW_C012

#ifdef LINK_REALLY_SLOW_C012
#define out		outb_p
#define in		inb_p
#define LINK_SLOW_DOWN	"outb %%al,$0x80\n\t"
#else
#define out		outb
#define in		inb
#define LINK_SLOW_DOWN	""
#endif

#ifdef LINK_DEBUG
#define DEB(x)		x
#define PRINTK(f, a...)	printk(f, ##a)
#define NO_DEB(x)
#else
#define DEB(x)
#define PRINTK(f, a...)
#define NO_DEB(x)	x
#endif

#ifdef LINK_DEBUG_MORE
#define DEB_MORE(x) x
#else
#define DEB_MORE(x)
#endif

#if !defined(SYMBOL_NAME_STR)
#if defined(__ELF__)
#define SYMBOL_NAME_STR(X) #X
#else
#define SYMBOL_NAME_STR(X) "_"#X
#endif
#endif


static struct link_struct link_table[LINK_NO] =
	{
		 { LINK_READ_ABORT | LINK_WRITE_ABORT, 0, 0, 0, 0,
		 LINK_INIT_READ_TIMEOUT, LINK_INIT_WRITE_TIMEOUT, LINK_B004 },
		 { LINK_READ_ABORT | LINK_WRITE_ABORT, 0, 0, 0, 0,
		 LINK_INIT_READ_TIMEOUT, LINK_INIT_WRITE_TIMEOUT, LINK_B004 },
		 { LINK_READ_ABORT | LINK_WRITE_ABORT, 0, 0, 0, 0,
		 LINK_INIT_READ_TIMEOUT, LINK_INIT_WRITE_TIMEOUT, LINK_B004 },
	};

static unsigned char link_devices = 0;	/* number of devices dectected */
static unsigned char use_schedule = 0;	/* don't use schedule() in init() */

/*
 * At these addresses the driver will search for B004-compatible boards
 */
static const short link_base_addresses[] = { 0x150, 0x200, 0x250, 0x300, 0 };

static const char *link_name(dev_t minor)
{
	static char name[] = LINK_NAME"?";

	if (minor < link_devices)
	{
		name[4] = minor + '0';
		return name;
	}
	else
		return (const char*)0;
}

static void link_delay(void)
{
	if ( use_schedule )
	{
		PRINTK("link_delay: Using schedule()\n");
		current->state = TASK_INTERRUPTIBLE;
		current->timeout = jiffies + 10;   /* 100ms */
		schedule();
	}
	else
	{		/* only for detection in link_init() */
		PRINTK("link_delay: Using idle loop\n");
		asm(	"movl $10000, %%ecx\n"
			"1:\tloop 1b" : : : "ecx" );
			
	}

}


/****************************************************************************
 *
 * static void link_reset():
 *
 * reset transputer network.
 *
 ****************************************************************************/

static void link_reset( const int minor )
{

	PRINTK("LINK(%d) resetting transputer.\n", minor);

	out( LINK_DEASSERT_ANALYSE, LINK_ANALYSE(minor));
	link_delay();

	out( LINK_DEASSERT_RESET, LINK_RESET(minor));
	link_delay();

	out( LINK_ASSERT_RESET, LINK_RESET(minor));
	link_delay();

	out( LINK_DEASSERT_RESET, LINK_RESET(minor));
	link_delay();

}

/****************************************************************************
 *
 * static void link_analyse():
 *
 * switch transputer network to analyse mode.
 *
 ****************************************************************************/

static void link_analyse( const int minor )
{

	PRINTK("LINK(%d) switching transputer to analyse mode.\n", minor);

	out( LINK_DEASSERT_ANALYSE, LINK_ANALYSE(minor));
	link_delay();

	out( LINK_ASSERT_ANALYSE, LINK_ANALYSE(minor));
	link_delay();

	out( LINK_ASSERT_RESET, LINK_RESET(minor));
	link_delay();

	out( LINK_DEASSERT_RESET, LINK_RESET(minor));
	link_delay();

	out( LINK_DEASSERT_ANALYSE, LINK_ANALYSE(minor));
	link_delay();


} /* link_analyse() */


/****************************************************************************
 *
 * static int link_read() - read bytes from the link interface.
 *
 * At first, the driver checks if the link-interface is ready to send a byte
 * to the PC. If not, this check is repeated up to LINK_MAXTRY times.
 * If the link-interface is not ready after this loop, the driver sleeps for
 * LINK_TIME jiffies and then checks the link-interface again.
 * If the interface is still not ready, the driver sleeps for LINK_READ_TIMEOUT
 * jiffies. Then it checks the interface again. If it is not ready and
 * the flag LINK_READ_ABORT is set, this function returns with the number
 * of bytes read or with an error if no byte was read.
 * If the LINK_READ_ABORT flag is not set, the driver sleeps again for
 * LINK_READ_ABORT jiffies and so on.
 *
 * By default, LINK_READ_ABORT is not set.
 *
 *****************************************************************************/

static int link_read(struct inode * inode, struct file * file, char * buf, int count)
{

	const unsigned int minor = MINOR(inode->i_rdev);
	unsigned int read_sleep = LINK_START_SLEEP;
	unsigned int read_wait  = 0;
	unsigned int read_inc = 0;
	DEB( unsigned int link_total_bytes_read = 0; )
#ifndef LINK_USE_ASM
	char 	*temp = buf;
	int	l_count;
	char	buffer[LINK_MAX_BYTES];
	int	end;
	int	i;
#else
	register char 	*temp __asm__("di") = buf;
#endif

	DEB_MORE(printk("LINK reading %d bytes from link %d.\n", count, minor);)

	if ( count < 0)
	{
		PRINTK("LINK: invalid argument for reading: count = %d.\n", count);
		return -EINVAL;
	}

#ifdef LINK_USE_ASM
	__asm__("cld": :);

	while ( 1 )
	{
		/*
		 * This is very ugly! For a description of this algorithm
		 * see the C-version below.
		 */
		__asm__(
			"movl	%4,%0\n\t"	/* we have to transfer %0 bytes */
			"push	%%es\n\t"
			"movw	%%fs,%%ax\n\t"
			"movw	%%ax,%%es\n"
		"9:	movl	%0,%4\n\t"
			"movb	$2,%%ah\n\t"
			"cmpl	%7,%4\n\t"
			"jc	6f\n\t"
			"movl	%7,%4\n"
		"6:	movl	%4,%%esi\n"
		"5:	inb	%w2,%%al\n\t"	/* ready for reading? */
			LINK_SLOW_DOWN
			"andb	%6,%%al\n\t"
			"jnz	1f\n\t"
			"pushl	%4\n\t"		/* No */
			"movl	%5,%4\n"
		"3:	inb	%w2,%%al\n\t"
			LINK_SLOW_DOWN
			"andb	%6,%%al\n\t"
			"jnz	2f\n\t"
			"decl	%4\n\t"
			"jnz	3b\n\t"
			"popl	%4\n\t"
			"jmp	7f\n"		/* timeout */
		"2:	popl	%4\n"
		"1:	subb	%%ah,%b2\n\t"	/* set dx to IDR */
			"insb\n\t"		/* read one byte */
			LINK_SLOW_DOWN
			"addb	%%ah,%b2\n\t"	/* set it back to ISR */
			"decl	%4\n\t"		/* decrement bytes to go */
			"jnz	5b\n\t"
			"subl	%%esi,%0\n\t"
			"jz	8f\n\t"		/* all transfered */
			"pop	%%es\n\t"	/* after transfering LINK_MAX_BYTES */
			"pushl	%%edx\n\t"	/* it's time to call the */
			"call "	SYMBOL_NAME_STR(schedule)"\n\t"	/* scheduler */
			"popl	%%edx\n\t"
			"push	%%es\n\t"
			"movw	%%fs,%%ax\n\t"
			"movw	%%ax,%%es\n\t"
			"jmp	9b\n"
		"7:	subl	%4,%%esi\n\t"
			"subl	%%esi,%0\n"
		"8:	pop	%%es"

			:"=b" (count),
			 "=D" (temp)
			:"d" ((unsigned short) LINK_ISR(minor)),
			 "1" (temp),
			 "c" (count),
			 "i" (LINK_MAXTRY),
			 "i" (LINK_READBYTE),
			 "i" (LINK_MAX_BYTES)
			:"ax", "di", "cx", "dx", "si");
		DEB(link_total_bytes_read += temp - buf);

		if (!count)
			return temp - buf;
#else
	while ( count )
	{
		end = (LINK_MAX_BYTES > count) ? count : LINK_MAX_BYTES;
		l_count = 0;
		do
		{
			if (!in(LINK_ISR(minor)) & LINK_READBYTE)
			{
				i = LINK_MAXTRY;
				while (!(in(LINK_ISR(minor)) & LINK_READBYTE) && i)
					i--;
				if (!i)
					break;
			}
			buffer[l_count++] = (char) in(LINK_IDR(minor));
			DEB(link_total_bytes_read++;)
		} while (--end);

		if (l_count)
		{
			memcpy_tofs(temp, buffer, l_count);
			count -= l_count;
			temp += l_count;
			if ( need_resched ) 
				schedule();
		}
#endif /* LINK_USE_ASM */
		else while ( !( in(LINK_ISR(minor)) & LINK_READBYTE ) )
		{
			/* check for signals before going to sleep. */
			if (current->signal & ~current->blocked)
				return temp - buf ? : -EINTR;

			PRINTK("LINK(%d) sleeping at %d bytes for %d jiffies (reading).\n",
					minor, link_total_bytes_read, read_sleep );

			current->state = TASK_INTERRUPTIBLE;
			current->timeout = jiffies + read_sleep;
			schedule();

			read_wait += read_sleep;
			read_inc = (read_inc + 1) % LINK_INC;
			read_sleep += (read_sleep < LINK_MAX_SLEEP && !read_inc);

			if ( read_wait > LINK_READ_TIMEOUT(minor)
				&& (LINK_F(minor) & LINK_READ_ABORT) )
			{
				PRINTK("LINK: timeout reading from link %d.\n", minor);
				return temp - buf;
			}
				
		}
	}
	DEB_MORE(printk("LINK reading successful.\n");)

	return temp - buf;

} /* LINK_read() */




/****************************************************************************
 *
 * static int link_write() - write to the link interface.
 *
 *****************************************************************************/

static int link_write(struct inode * inode, struct file * file, char * buf, int count)
{
	const unsigned int minor = MINOR(inode->i_rdev);
	unsigned int write_sleep = LINK_START_SLEEP;
	unsigned int write_wait  = 0;
	unsigned int write_inc = 0;
	DEB( unsigned int link_total_bytes_written = 0; )
#ifndef LINK_USE_ASM
	char	*temp = buf;
	int	l_count;
	int	end;
	char	buffer[LINK_MAX_BYTES];
	int	i;
#else
	register char	*temp __asm__("si") = buf;
#endif

	DEB_MORE(printk("LINK writing %d bytes to link %d.\n", count, minor);)

	if ( count < 0)
	{
		PRINTK("LINK invalid argument for writing: count = %d.\n", count);
		return -EINVAL;
	}

#ifdef LINK_USE_ASM
		__asm__("cld": :);

	while ( 1 )
	{
		/*
		 * this is very ugly!
		 */
		__asm__(
			"movl	%4,%0\n\t"
			"push	%%ds\n\t"
			"push	%%fs\n\t"
			"movw	%%fs,%%ax\n\t"
			"movw	%%ds,%w4\n\t"
			"movw	%%ax,%%ds\n\t"
			"movw	%w4,%%fs\n"
		"9:	movl	%0,%4\n\t"
			"movb	$2,%%ah\n"
			"cmpl	%7,%4\n\t"
			"jc	6f\n\t"
			"movl	%7,%4\n"
		"6:	movl	%4,%%edi\n"
		"5:	inb	%w2, %%al\n\t"
			LINK_SLOW_DOWN
			"andb	%6,%%al\n\t"
			"jnz	1f\n\t"
			"pushl	%4\n\t"
			"movl	%5,%4\n"
		"3:	inb	%w2, %%al\n\t"
			LINK_SLOW_DOWN
			"andb	%6,%%al\n\t"
			"jnz	2f\n\t"
			"decl	%4\n\t"
			"jnz	3b\n\t"
			"popl	%4\n\t"
			"jmp	7f\n"
		"2:	popl	%4\n"
		"1:	subb	%%ah,%b2\n\t"
			"outsb\n\t"
			LINK_SLOW_DOWN
			"addb	%%ah,%b2\n\t"
			"decl	%4\n\t"
			"jnz	5b\n\t"
			"subl	%%edi,%0\n\t"
			"jz	8f\n\t"
			"pop	%%fs\n\t"	/* after transfering */
			"pop	%%ds\n\t"	/* LINK_MAX_BYTES it is time */
			"pushl	%%edx\n\t"	/* call the scheduler */
			"call "	SYMBOL_NAME_STR(schedule)"\n\t"	/* scheduler */
			"popl	%%edx\n\t"
			"push	%%ds\n\t"
			"push	%%fs\n\t"
			"movw	%%fs,%%ax\n\t"
			"movw	%%ds,%w4\n\t"
			"movw	%%ax,%%ds\n\t"
			"movw	%w4,%%fs\n\t"
			"jmp	9b\n"
		"7:	subl	%4,%%edi\n\t"
			"subl	%%edi,%0\n"
		"8:	pop	%%fs\n\t"
			"pop	%%ds"

			:"=b" (count),
			 "=S" (temp)
			:"d" ((unsigned short) LINK_OSR(minor)),
			 "1" (temp),
			 "c" (count),
			 "i" (LINK_MAXTRY),
			 "i" (LINK_WRITEBYTE),
			 "i" (LINK_MAX_BYTES)
			:"ax", "si", "dx", "cx", "di");
		DEB(link_total_bytes_written += temp - buf);

		if (!count)
			return temp - buf;
#else
	while ( count )
	{
		end = (LINK_MAX_BYTES > count) ? count : LINK_MAX_BYTES;
		memcpy_fromfs(buffer, temp, end);
		l_count = 0;
		do
		{
			if (!in(LINK_OSR(minor)) & LINK_WRITEBYTE)
			{
				i = LINK_MAXTRY;
				while (!(in(LINK_OSR(minor)) & LINK_WRITEBYTE) && i)
					i--;
				if (!i)
					break;
			}
			out(buffer[l_count++], LINK_ODR(minor));
			DEB(link_total_bytes_written++;)
		} while (--end);

		if (l_count)
		{
			temp += l_count;
			count -= l_count;

			if ( need_resched )
				schedule();
		}
#endif	/* LINK_USE_ASM */
		else while ( !( in(LINK_OSR(minor)) & LINK_WRITEBYTE ) )
		{
			/* check for signals before going to sleep. */
			if (current->signal & ~current->blocked)
				return temp - buf ? : -EINTR;

			PRINTK("LINK(%d) sleeping at %d bytes for %d jiffies (writing).\n",
					minor, link_total_bytes_written, write_sleep );

			current->state = TASK_INTERRUPTIBLE;
			current->timeout = jiffies + write_sleep;
			schedule();

			write_wait += write_sleep;
			write_inc = (write_inc + 1) % LINK_INC;
			write_sleep += (write_sleep < LINK_MAX_SLEEP && !write_inc);

			if ( write_wait > LINK_WRITE_TIMEOUT(minor)
				&& (LINK_F(minor) & LINK_WRITE_ABORT) )
			{
				PRINTK("LINK: timeout writing to link %d.\n", minor);
				return temp - buf;
			}
				
		}
	}

	DEB_MORE(printk("LINK writing successful.\n");)
	return temp-buf;
} /* link_write() */

/****************************************************************************
 *
 * static int link_lseek()
 *
 ***************************************************************************/
static int link_lseek(struct inode * inode, struct file * file,
		    off_t offset, int origin)
{
	return -ESPIPE;
}

/****************************************************************************
 *
 * static int link_open()
 *
 * open the link-device.
 *
 ***************************************************************************/

static int link_open(struct inode * inode, struct file * file)
{
	const unsigned int minor = MINOR(inode->i_rdev);

	if (minor >= link_devices)
	{
		PRINTK("LINK not opened, minor device number >= %d.\n", link_devices);
		return -ENODEV;
	}

	if (LINK_F(minor) & LINK_BUSY)
	{
		PRINTK("LINK not opened, LINK-board busy (minor = %d).\n", minor);
		return -EBUSY;
	}

	LINK_F(minor) |= LINK_BUSY;

#ifdef MODULE
	MOD_INC_USE_COUNT;
#endif

	PRINTK( "LINK(%d) opened.\n", minor);
	return 0;

} /* link_open() */


/****************************************************************************
 *
 * static int link_release()
 *
 * close the link device.
 *
 ****************************************************************************/

static void link_release(struct inode * inode, struct file * file)
{
	const unsigned int minor = MINOR(inode->i_rdev);

	if (minor >= link_devices)
	{
		PRINTK("LINK not released, minor device number >= %d.\n", link_devices);
		return;
	}

	LINK_F(minor) &= ~LINK_BUSY;

#ifdef MODULE
	MOD_DEC_USE_COUNT;
#endif

	PRINTK("LINK(%d) released.\n", minor);

} /* link_release() */


/****************************************************************************
 *
 * static int link_ioctl()
 *
 * This function performs the various ioctl() functions: resetting the
 * transputer, switching to analyse-mode, testing the status, changing
 * timeouts etc.
 *
 *****************************************************************************/

static int link_ioctl(struct inode *inode, struct file *file,
		    unsigned int cmd, unsigned long arg)
{
	const unsigned int minor = MINOR(inode->i_rdev);
	int result = arg;

	PRINTK("LINK(%d) ioctl, cmd: 0x%x, arg: 0x%x.\n", minor, cmd, arg);

	if (minor >= link_devices || !(LINK_F(minor) & LINK_BUSY) )
	{
		DEB(
			if (minor >= link_devices)
				printk("LINK ioctl exit, minor >= %d.\n", link_devices );
			else
				printk("LINK ioctl exit, device not opened.\n" );
		)
		return -ENODEV;
	}

	switch (cmd)
	{
		case LINKRESET:		/* reset transputer */
			link_reset(minor);
			break;
		case LINKWRITEABLE:	/* can we write a byte to the C012 ? */
	 		return ( ( in(LINK_OSR(minor)) & LINK_WRITEBYTE) != 0 ); 
		case LINKREADABLE:	/* can we read a byte from C012 ? */
	 		return ( ( in(LINK_ISR(minor)) & LINK_READBYTE) != 0 ); 
		case LINKANALYSE:	/* switch transputer to analyse mode */
			link_analyse(minor);
			break;
		case LINKERROR:		/* test error-flag */
			return ( in(LINK_BASE(minor) + LINK_ERROR_OFFSET) & LINK_TEST_ERROR) ? 0 : 1;
		case LINKREADTIMEOUT:	/* set timeout for reading */
			result = LINK_READ_TIMEOUT(minor);
			LINK_READ_TIMEOUT(minor) = arg;
			break;
		case LINKWRITETIMEOUT:	/* set timeout for writing */
			result = LINK_WRITE_TIMEOUT(minor);
			LINK_WRITE_TIMEOUT(minor) = arg;
			break;
		case LINKREADABORT:	/* abort after a timeout ? */
			if ( arg )
				LINK_F(minor) |= LINK_READ_ABORT;
			else
				LINK_F(minor) &= ~LINK_READ_ABORT;
			break;
		case LINKWRITEABORT:	/* abort after a timeout ? */
			if ( arg )
				LINK_F(minor) |= LINK_WRITE_ABORT;
			else
				LINK_F(minor) &= ~LINK_WRITE_ABORT;
			break;
		default: result = -EINVAL;
	}

	PRINTK("LINK(%d) ioctl done.\n\n", minor);

	return result;

} /* link_ioctl() */


static struct file_operations link_fops = {
	link_lseek,
	link_read,
	link_write,
	NULL,		/* link_readdir */
	NULL,		/* link_select */
	link_ioctl,
	NULL,		/* link_mmap */
	link_open,
	link_release
};

/****************************************************************************
 *
 * long link_init()
 *
 * This function initializes the driver. It tries to detect the hardware
 * and sets up all relevant data structures.
 *
 ****************************************************************************/

long link_init(long kmem_start)
{
	unsigned int test, i;

	if ( register_chrdev( LINK_MAJOR, LINK_NAME, &link_fops ) )
	{
		printk("link_init: unable to get major %d for link interface.\n",
			LINK_MAJOR );
		return kmem_start;
	}

	/*
	   After a reset it should be possible to write a byte to
	   the LINK. So let's do a reset and then test the output status
	   register
	*/
	for (test = 0; link_base_addresses[test] &&
					link_devices < LINK_NO; test++)
	{
		link_delay();
		LINK_BASE((int) link_devices) = link_base_addresses[test];
		LINK_ODR((int) link_devices) = LINK_BASE((int) link_devices) + LINK_ODR_OFFSET;
		LINK_ISR((int) link_devices) = LINK_BASE((int) link_devices) + LINK_ISR_OFFSET;
		LINK_OSR((int) link_devices) = LINK_BASE((int) link_devices) + LINK_OSR_OFFSET;
		link_reset(link_devices);
		link_delay();

		for (i = 0; i < LINK_MAXTRY; i++)
			if ( in(LINK_OSR((int) link_devices)) == LINK_WRITEBYTE)
			{
				out(LINK_BASE((int) link_devices) + B008_INT_OFFSET, 0);
				link_delay();
				if ((in(LINK_BASE((int) link_devices) + B008_INT_OFFSET) & 0x0f) == 0)
					LINK_BOARDTYPE((int) link_devices) = LINK_B008;
				else
					LINK_BOARDTYPE((int) link_devices) = LINK_B004;
				printk("link%d at 0x0%x (polling) is a B00%s\n",
					link_devices,LINK_IDR((int) link_devices),
					LINK_BOARDTYPE((int) link_devices) == LINK_B004 ? "4" : "8");
				request_region(LINK_IDR((int) link_devices), 
					LINK_BOARDTYPE((int) link_devices) == LINK_B004 ? B004_IO_SIZE : B008_IO_SIZE,
					LINK_NAME);
				link_devices++;
				break;
			}

		}

	use_schedule = 1;

	if (!link_devices)
		printk("link: no interfaces found.\n");

	return kmem_start;

} /* link_init() */

#ifdef MODULE
int init_module(void)
{
	long dummy = 0;

	dummy = link_init(dummy);

	return 0;
}

void cleanup_module(void)
{
	int i;
	if (MOD_IN_USE)
		printk("link: device busy, remove delayed.\n");
	else
	{
		unregister_chrdev( LINK_MAJOR, LINK_NAME );
		for (i = 0; i < link_devices; i++)
			release_region(LINK_IDR(i), LINK_BOARDTYPE(i) ==
				LINK_B004 ? B004_IO_SIZE : B008_IO_SIZE);
	}
}
#endif
